// UBC CPSC 314 (2015W2) -- P1
// HAVE FUN!!! :)

// ASSIGNMENT-SPECIFIC API EXTENSION
THREE.Object3D.prototype.setMatrix = function(a) {
  if (isNaN(a.determinant()))
    console.error("det is NaN!", a);

  this.matrix=a;
  this.matrix.decompose(this.position,this.quaternion,this.scale);
}

// SETUP RENDERER & SCENE
var texture, sphere;
var canvas = document.getElementById('canvas');
var scene = new THREE.Scene();
var renderer = new THREE.WebGLRenderer({ antialias: true });
renderer.shadowMapEnabled = true;
//renderer.shadowMapType = THREE.PCFSoftShadowMap;
renderer.setClearColor(0xFFFFFF); // white background colour
canvas.appendChild(renderer.domElement);

// SETUP CAMERA
var camera = new THREE.PerspectiveCamera(30,1,0.1,1000); // view angle, aspect ratio, near, far
camera.position.set(45,20,40);
camera.lookAt(scene.position);
scene.add(camera);

var cubeCamera = new THREE.CubeCamera(50, 2000, 1024);
cubeCamera.renderTarget.minFilter = THREE.LinearMipMapLinearFilter;
scene.add(cubeCamera);

// SETUP LIGHTS
var ambientLight = new THREE.AmbientLight(0x80aaff);
scene.add(ambientLight);

var directionalLight = new THREE.DirectionalLight(0x884411);
directionalLight.position.set(10, 20, 10);
directionalLight.lookAt(new THREE.Vector3());
directionalLight.castShadow = true;
directionalLight.shadowDarkness = 0.5;
directionalLight.shadowMapWidth = directionalLight.shadowMapHeight = 512;
directionalLight.shadowCameraNear = 10;
directionalLight.shadowCameraFar = 30;
directionalLight.shadowBias = 1e-5;
directionalLight.shadowCameraTop = directionalLight.shadowCameraRight = 10;
directionalLight.shadowCameraBottom = directionalLight.shadowCameraLeft = -10;
directionalLight.shadowMapHeight = directionalLight.shadowMapWidth = 1024;

//directionalLight.shadowCameraVisible = true;
scene.add(directionalLight);

var pointLight = new THREE.PointLight(0x0099ff, 1, 23);
pointLight.position.set(10, 12, 2);
scene.add(pointLight);

// SETUP ORBIT CONTROLS OF THE CAMERA
var controls = new THREE.OrbitControls(camera);

// ADAPT TO WINDOW RESIZE
function resize() {
  renderer.setSize(window.innerWidth,window.innerHeight);
  camera.aspect = window.innerWidth/window.innerHeight;
  camera.updateProjectionMatrix();
}

// EVENT LISTENER RESIZE
window.addEventListener('resize',resize);
resize();

//SCROLLBAR FUNCTION DISABLE
window.onscroll = function () {
     window.scrollTo(0,0);
   }

// SETUP HELPER GRID
// Note: Press Z to show/hide
var gridGeometry = new THREE.Geometry();
var i;
for(i=-50;i<51;i+=2) {
    gridGeometry.vertices.push( new THREE.Vector3(i,0,-50));
    gridGeometry.vertices.push( new THREE.Vector3(i,0,50));
    gridGeometry.vertices.push( new THREE.Vector3(-50,0,i));
    gridGeometry.vertices.push( new THREE.Vector3(50,0,i));
}

var gridMaterial = new THREE.LineBasicMaterial({color:0xBBBBBB});
var grid = new THREE.Line(gridGeometry,gridMaterial,THREE.LinePieces);

/////////////////////////////////
//   YOUR WORK STARTS BELOW    //
/////////////////////////////////

// MATERIALS
// Note: Feel free to be creative with this!
//var normalMaterial = new THREE.MeshNormalMaterial();

var normalMaterial = new THREE.MeshPhongMaterial({
  color: 0xffcc00,
  envMap: cubeCamera.renderTarget,
  reflectivity: 0.6,
  refractionRatio: 0.8
});

function makeCube() {
  var unitCube = new THREE.BoxGeometry(1,1,1);
  return unitCube;
}

// Create a matrix given components
function mat(n11, n12, n13, n14,
             n21, n22, n23, n24,
             n31, n32, n33, n34,
             n41, n42, n43, n44) {
  return new THREE.Matrix4().set(n11, n12, n13, n14,
                                 n21, n22, n23, n24,
                                 n31, n32, n33, n34,
                                 n41, n42, n43, n44);
}

function Translate(a, b, c) {
  return mat(1, 0, 0, a,
             0, 1, 0, b,
             0, 0, 1, c,
             0, 0, 0, 1);
}

function Scale(a, b, c) {
  return mat(a, 0, 0, 0,
             0, b, 0, 0,
             0, 0, c, 0,
             0, 0, 0, 1);
}

function RotX(theta, scale) {
  if (arguments.length == 1) scale = 1;
  var c = scale*Math.cos(theta), s = scale*Math.sin(theta), x = scale;
  return mat(x, 0,  0, 0,
             0, c, -s, 0,
             0, s,  c, 0,
             0, 0,  0, 1);
}

function RotY(theta, scale) {
  if (arguments.length == 1) scale = 1;
  var c = scale*Math.cos(theta), s = scale*Math.sin(theta), x = scale;
  return mat(c, 0, -s, 0,
             0, x,  0, 0,
             s, 0,  c, 0,
             0, 0,  0, 1);
}

function RotZ(theta, scale) {
  if (arguments.length == 1) scale = 1;
  var c = scale*Math.cos(theta), s = scale*Math.sin(theta), x = scale;
  return mat(c, -s, 0, 0,
             s,  c, 0, 0,
             0,  0, x, 0,
             0,  0, 0, 1);
}

function multMat(m1, m2) {
  return new THREE.Matrix4().multiplyMatrices(m1, m2);
}

function multMat3(m1, m2, m3) {
  return multMat(m1, multMat(m2, m3));
}

function multMat4(m1, m2, m3, m4) {
  return multMat(m1, multMat(m2, multMat(m3, m4)));
}

// Perform transformation t around a point (x, y, z)
function Around(x, y, z, r) {
  return multMat3(Translate(-x, -y, -z), r, Translate(x, y, z));
}

function deg(x) {
  return x * Math.PI / 180.0;
}

function clamp(x, a, b) {
  if (x < a) return a;
  if (x > b) return b;
  return x;
}

function lerp(t, a, b) {
  t = clamp(t, 0, 1);
  return (1-t)*a + t*b;
}

function unlerp(u, a, b) {
  u = clamp(u, a, b);
  return (u - a) / (b - a);
}

function pingpong(n) {
  return function(t, a, b) {
    var s = Math.sin(Math.PI * (n + 1/2) * t);
    return lerp(s*s, a, b);
  }
}
function wave(n) {
  return function(t, a, b) {
    var s = Math.sin(Math.PI * (n - 1/2) * t);
    return a + (b-a)*s;
  }
}

function MatStk(m0) {
  var N = 64;
  var matStk = new Array(N);
  var matIdx = 0;
  matStk[0] = m0 || mat(1, 0, 0, 0,
                        0, 1, 0, 0,
                        0, 0, 1, 0,
                        0, 0, 0, 1);

  function getSize() {
    return matIdx + 1;
  }

  function get() {
    return matStk[matIdx];
  }

  function push(m) {
    if (matIdx == N) throw new Error("Can't push.");
    var oldMat = matStk[matIdx];
    var newMat = new THREE.Matrix4().multiplyMatrices(oldMat, m);
    matStk[++matIdx] = newMat;
    return this;
  }

  function pop() {
    if (matIdx == 0) throw new Error("Can't pop.");
    matStk[matIdx] = null;
    matStk[--matIdx];
    return this;
  }

  this.get = get;
  this.push = push;
  this.pop = pop;
  this.getSize = getSize;
}

// GEOMETRY
var torsoGeometry = makeCube(), headGeometry = makeCube(),
    tailGeometry = makeCube(), noseGeometry = makeCube(),
    clawGeometry = makeCube(), legGeometry = makeCube(),
    midNoseGeometry = makeCube(), nailGeometry = makeCube();
var torsoScale = Scale(5, 5, 8),
    headScale = Scale(3, 3, 2),
    tailScale = Scale(0.8, 0.8, 1),
    clawScale = Scale(2, 1, 3),
    legScale = Scale(2, 1, 2),
    nailScale = Scale(0.5, 0.5, 1),
    noseScale = Scale(0.4, 0.4, 2.5),
    midNoseScale = Scale(0.35, 0.3, 2);

torsoGeometry.applyMatrix(torsoScale);
headGeometry.applyMatrix(headScale);
tailGeometry.applyMatrix(tailScale);
clawGeometry.applyMatrix(clawScale);
legGeometry.applyMatrix(legScale);
nailGeometry.applyMatrix(nailScale);
noseGeometry.applyMatrix(noseScale);
midNoseGeometry.applyMatrix(midNoseScale);

const NOSE_COUNT = 18;
const TAIL_PARTS = 15;
const NAIL_COUNT = 3;


function addObject(x) {
  x.castShadow = x.receiveShadow = true;
  return scene.add(x);
}

// TODO: SPECIFY THE REST OF YOUR STAR-NOSE MOLE'S GEOMETRY.
// Note: You will be using transformation matrices to set the shape.
// Note: You are not allowed to use the tools Three.js provides for
//       rotation, translation and scaling.
// Note: The torso has been done for you (but feel free to modify it!)
// Hint: Explicity declare new matrices using Matrix4().set

// CREATE BODY
var torso = new THREE.Mesh(torsoGeometry, normalMaterial),
    head = new THREE.Mesh(headGeometry, normalMaterial),
    clawR = new THREE.Mesh(clawGeometry, normalMaterial),
    clawL = new THREE.Mesh(clawGeometry, normalMaterial),
    legR = new THREE.Mesh(legGeometry, normalMaterial),
    legL = new THREE.Mesh(legGeometry, normalMaterial),
    tailParts = [],
    noseParts = [],
    midNoseParts = [];
    nailPartsR = [];
    nailPartsL = [];
    nailPartsBR = [];
    nailPartsBL = [];

// TODO 2 or 3 segment torso

addObject(torso);
addObject(head);
addObject(clawR);
addObject(clawL);
addObject(legR);
addObject(legL);

for (var i = 0; i < TAIL_PARTS; i++) {
  tailParts.push(new THREE.Mesh(tailGeometry, normalMaterial));
}
tailParts.forEach(function(p) { addObject(p); });

for (var i = 0; i < NOSE_COUNT; i++) {
  noseParts.push(new THREE.Mesh(noseGeometry, normalMaterial));
}

noseParts.forEach(function(p) { addObject(p); });

for (var i = 0; i < 4; i++) {
  midNoseParts.push(new THREE.Mesh(midNoseGeometry, normalMaterial));
}

midNoseParts.forEach(function(p) { addObject(p); });

for (var i = 0; i < NAIL_COUNT; i++) {
  nailPartsR.push(new THREE.Mesh(nailGeometry, normalMaterial));
}
nailPartsR.forEach(function(p) { addObject(p); });

for (var i = 0; i < NAIL_COUNT; i++) {
  nailPartsL.push(new THREE.Mesh(nailGeometry, normalMaterial));
}
nailPartsL.forEach(function(p) { addObject(p); });

for (var i = 0; i < NAIL_COUNT; i++) {
  nailPartsBR.push(new THREE.Mesh(nailGeometry, normalMaterial));
}
nailPartsBR.forEach(function(p) { addObject(p); });

for (var i = 0; i < NAIL_COUNT; i++) {
  nailPartsBL.push(new THREE.Mesh(nailGeometry, normalMaterial));
}
nailPartsBL.forEach(function(p) { addObject(p); });

// Hint: Hint: Add one piece of geometry at a time, then implement the motion for that part.
//             Then you can make sure your hierarchy still works properly after each step.

// MATRICES
// TODO: INITIALIZE THE REST OF YOUR MATRICES
// Note: Use of parent attribute is not allowed.
// Hint: Keep hierarchies in mind!
// Hint: Play around with the headTorsoMatrix values, what changes in the render? Why?
function updateMatrices() {
  var torsoMatrix = Translate(0, 2.5, 1);
  var headMatrix = Translate(0, 0, 5);
  var noseMatrix = Translate(0, 0, 2);
  var tailMatrix = Translate(0, 0, -3);


  var headRot = RotY(state.headAngle);
  var torsoRotZ = RotX(state.torsoAngle);

  var clawRMatrix = Translate(3.5, -2, 3.5);
  var clawLMatrix = Translate(-3.5, -2, 3.5);
  var legRMatrix = Translate(3.5, -2, -2.5);
  var legLMatrix = Translate(-3.5, -2, -2.5);
  var tailTrans = Translate(0, 0, -1);

  var nailRMatrix = Translate(-2, 0, 2);
  var nailLMatrix = Translate(-2, 0, 2);
  var nailBRMatrix = Translate(-1.5, 0, 1.5);
  var nailBLMatrix = Translate(-1.5, 0, 1.5);

  var nailSpreadR = Translate(1, 0, 0);
  var nailSpreadL = Translate(1, 0, 0);
  var nailSpreadBR = Translate(0.75, 0, 0);
  var nailSpreadBL = Translate(0.75, 0, 0);

  var clawRRot = RotX(state.clawRAngle);
  var clawLRot = RotX(state.clawLAngle)
  var legRRot = RotX(state.legRAngle);
  var legLRot = RotX(state.legLAngle);
  var nailRRot = Around(0.75, 0, 1., RotX(state.nailAngleR));
  var nailLRot = Around(-0.75, 0, 1., RotX(state.nailAngleL));
  var nailBRRot = Around(-2, 0, 1., RotX(state.nailAngleBR));
  var nailBLRot = Around(2, 0, 1., RotX(state.nailAngleBL));

  var tailDeltaScale = 0.9;
  var tailRot = multMat3(
      RotZ(-state.tailTilt),
      RotY(-state.tailAngle, tailDeltaScale),
      RotZ(state.tailTwist + state.tailTilt));
  var tailDelta = multMat(tailTrans, tailRot);

  var s = new MatStk(torsoMatrix);
  torso.setMatrix(s.push(torsoRotZ).get());

  // HEAD [
  s.push(headMatrix).push(headRot);
  head.setMatrix(s.get());

  // - NOSE
  s.push(noseMatrix);

  const noseSpread = 0.8,
        noseSpacing = 0.6,
        midNoseSpread = 0.5,
        midNoseSpacing = 0.25;

  midNoseParts.forEach(function(part, i) {
    var side = [-1, -1, 1, 1][i], sgn = [-1, 1, -1, 1][i];
    s.push(Translate(side * midNoseSpacing, sgn * midNoseSpread, 0));
    s.push(Around(0, 0, 1, RotX(-sgn * 0.8 * state.noseAngle)));
    part.setMatrix(s.get());
    s.pop().pop();
  });

  noseParts.forEach(function(part, i) {
    var j = Math.floor(i / 2), parity = i % 2;
    var sgn = parity ? -1 : 1;
    var phi = deg((-sgn * j / (0.5*NOSE_COUNT - 1)) * 180 + 180);
    var delta = 1;

    function tr(x) {
      return Translate(0, 0, x);
    }

    var cosine = noseSpread * Math.cos(phi),
        sine = noseSpread * Math.sin(phi);
    s.push(Translate(noseSpacing * sgn + sine, -cosine, 0));
    s.push(RotZ(phi));
    s.push(Around(0, 0, delta, RotX(state.noseAngle)));

    part.setMatrix(s.get());

    s.pop().pop().pop();
  });
  s.pop();

  s.pop().pop();
  // ] HEAD

  // CLAW Right
  clawR.setMatrix(s.push(clawRMatrix).push(clawRRot).get());
  s.push(nailRMatrix).push(nailRRot);
  nailPartsR.forEach(function(part) {part.setMatrix(s.push(nailSpreadR).get()); });
  nailPartsR.forEach(function(part) { s.pop(); });
  s.pop().pop();
  s.pop().pop();

  // CLAW Left
  clawL.setMatrix(s.push(clawLMatrix).push(clawLRot).get());
  s.push(nailLMatrix).push(nailLRot)
  nailPartsL.forEach(function(part) {part.setMatrix(s.push(nailSpreadL).get()); });
  nailPartsL.forEach(function(part) { s.pop(); });
  s.pop().pop();
  s.pop().pop();

  // LEG Right
  legR.setMatrix(s.push(legRMatrix).push(legRRot).get());
  s.push(nailBRMatrix).push(nailBRRot)
  nailPartsBR.forEach(function(part) {part.setMatrix(s.push(nailSpreadBR).get()); });
  nailPartsBR.forEach(function(part) { s.pop(); });
  s.pop().pop();
  s.pop().pop();

  // LEG Left
  legL.setMatrix(s.push(legLMatrix).push(legLRot).get());
  s.push(nailBLMatrix).push(nailBLRot)
  nailPartsBL.forEach(function(part) {part.setMatrix(s.push(nailSpreadBL).get()); });
  nailPartsBL.forEach(function(part) { s.pop(); });
  s.pop().pop();
  s.pop().pop();

  // TAIL
  s.push(tailMatrix);
  tailParts.forEach(function(part) {
    var m1 = s.push(tailDelta).get();
    part.setMatrix(m1);
  });
  tailParts.forEach(function(part) { s.pop(); });
  s.pop();

  // NAILS Front Right
 /* s.push(nailRMatrix).push(nailRRot);
  nailPartsR.forEach(function(part) {part.setMatrix(s.push(nailSpreadR).get()); });
  nailPartsR.forEach(function(part) { s.pop(); });
  s.pop().pop();*/

  // NAILS Front Left
  /*s.push(nailLMatrix).push(nailLRot)
  nailPartsL.forEach(function(part) {part.setMatrix(s.push(nailSpreadL).get()); });
  nailPartsL.forEach(function(part) { s.pop(); });
  s.pop().pop();*/

  // NAILS Back Right
  /*s.push(nailBRMatrix).push(nailBRRot)
  nailPartsBR.forEach(function(part) {part.setMatrix(s.push(nailSpreadBR).get()); });
  nailPartsBR.forEach(function(part) { s.pop(); });
  s.pop().pop();*/

  //NAILS Back Left
  /*s.push(nailBLMatrix).push(nailBLRot)
  nailPartsBL.forEach(function(part) {part.setMatrix(s.push(nailSpreadBL).get()); });
  nailPartsBL.forEach(function(part) { s.pop(); });
  s.pop().pop();*/
}



// APPLY DIFFERENT JUMP CUTS/ANIMATIONS TO DIFFERNET KEYS
// Note: The start of "U" animation has been done for you, you must implement the hiearchy and jumpcut.
// Hint: There are other ways to manipulate and grab clock values!!
// Hint: Check THREE.js clock documenation for ideas.
// Hint: It may help to start with a jumpcut and implement the animation after.
// Hint: Where is updateBody() called?
var clock = new THREE.Clock(true);

var jumpcut = false;

// Array of all running animations, higher priority animations are on the left
const Anims = [];

// The state variables for the mole. Animations will change it.
const state = {};

// Does object x has property k?
function hasKey(x, k) {
  return Object.prototype.hasOwnProperty.call(x, k);
}

function forEachState(st, f) {
  if (typeof st !== "object") {
    console.error("not an object!", st);
    throw new Error("not an object!");
  }

  var st1 = {};
  for (var k in st) {
    if (hasKey(st, k)) {
      var y = f(k, st[k]);
      if (typeof y !== "undefined")
        st1[k] = y;
    }
  }
  return st1;
}

function copyState(st, st0) {
  return forEachState(st, function(k, v) { st0[k] = v; });
}

// Given animation declaration and time, write interpolated state into a state
// variable
function evalState(start, end, fs, t, st) {
  forEachState(start, function(k, _) {
    var f = fs[k] || lerp;
    var v = f(t, start[k], end[k]);
    st[k] = v;
  });
}

// State declarations
// variables to describe the animal's state
const initialState = {
  torsoAngle: 0.0,
  headAngle: 0.0,
  noseAngle: deg(20.0),
  tailAngle: 0.0,
  tailTwist: 0.0,
  tailTilt: deg(10.0),
  clawRAngle: 0.0,
  clawLAngle: 0.0,
  legRAngle: 0.0,
  legLAngle: 0.0,
  nailAngleR: 0.0,
  nailAngleL: 0.0,
  nailAngleBR: 0.0,
  nailAngleBL: 0.0
};

// animations
const Animations = {
  'tiltUp': {
    start: {torsoAngle: 0},
    end: {torsoAngle: deg(-45)},
    // funcs: {torsoAngle: pingpong(3)},
    key: 'U'
  },
  'tiltDown': {
    start: {torsoAngle: 0},
    end: {torsoAngle: deg(45)},
    key: 'E'
  },
  'headLeft': {
    start: {headAngle: 0},
    end: {headAngle: deg(-30)},
    key: 'G'
  },
  'headRight': {
    start: {headAngle: 0},
    end: {headAngle: deg(30)},
    key: 'H'
  },
  'tailLeft': {
    start: {tailTilt: deg(0), tailTwist: deg(0),
       tailAngle: deg(0)},
    end: {tailTilt: deg(-20), tailTwist: deg(3),
       tailAngle: deg(-10)},
    key: 'V'
  },
  'tailRight': {
    start: {tailTilt: deg(0), tailTwist: deg(0),
       tailAngle: deg(0)},
    end: {tailTilt: deg(20), tailTwist: deg(-3),
       tailAngle: deg(10)},
    key: 'T'
  },
  'tailWag': {
    states : [

    {tailTilt: deg(0), tailTwist: deg(0), tailAngle: deg(0)},

    {tailTilt: deg(-0.5), tailTwist: deg(3), tailAngle: deg(0)},

    {tailTilt: deg(0.5), tailTwist: deg(-3), tailAngle: deg(0)},
    ],
    length: 3,
    funcs: {tailAngle: function(t) { return 0.3*Math.sin(2*Math.PI*t); }},
    key: 'W'
  },
  'noseFanOut': {
    start: {noseAngle: initialState.noseAngle},
    end: {noseAngle: deg(60)},
    key: 'N'
  },
  'swimOut' : {
    states : [
      {tailAngle: deg(0), noseAngle: initialState.noseAngle, headAngle: deg(0),
       clawLAngle: deg(0), clawRAngle: deg(0),
       nailAngleL: deg(0), nailAngleR: deg(0),
       nailAngleBL: deg(0), nailAngleBR: deg(0),
       legLAngle: deg(0), legRAngle: deg(0)},

      {tailAngle: deg(-6),  noseAngle: deg(60), headAngle: deg(40),
       clawLAngle: deg(0), clawRAngle: deg(60),
       nailAngleL: deg(0), nailAngleR: deg(60),
       nailAngleBL: deg(60), nailAngleBR: deg(0),
       legLAngle: deg(60), legRAngle: deg(0)},

      {tailAngle: deg(6),  noseAngle: deg(60), headAngle: deg(-40),
        clawLAngle: deg(60), clawRAngle: deg(0),
        nailAngleL: deg(60), nailAngleR: deg(0),
        nailAngleBL: deg(0), nailAngleBR: deg(60),
       legLAngle: deg(0), legRAngle: deg(50)}
    ],
    length: 0.5,
    key: 'S'
  },
  'dig': {
    states: [
      {clawLAngle: deg(0), clawRAngle: deg(0),
       nailAngleL: deg(0), nailAngleR: deg(0),
      },

      {clawLAngle: deg(60), clawRAngle: deg(60),
       nailAngleL: deg(90), nailAngleR: deg(90),
     },
    ],
    key: 'D'
  },
  'test': {
    states: [
      {noseAngle: initialState.noseAngle,
       tailTilt: deg(20), tailTwist: deg(3),
       headAngle: deg(0), tailAngle: deg(0),
       clawLAngle: deg(0), clawRAngle: deg(0),
       legLAngle: deg(0), legRAngle: deg(0)},

      {noseAngle: deg(60), headAngle: deg(20), tailAngle: deg(-5),
       tailTilt: deg(-20), tailTwist: deg(-10),
       clawLAngle: deg(60), clawRAngle: deg(0),
       legLAngle: deg(0), legRAngle: deg(60)},

      {noseAngle: deg(0), headAngle: deg(-50), tailAngle: deg(-15),
       tailTilt: deg(10), tailTwist: deg(-20),
       clawLAngle: deg(0), clawRAngle: deg(60),
       legLAngle: deg(60), legRAngle: deg(0)},

      {noseAngle: deg(80), headAngle: deg(70), tailAngle: deg(25),
       tailTilt: deg(-20), tailTwist: deg(5),
       clawLAngle: deg(60), clawRAngle: deg(0),
       legLAngle: deg(0), legRAngle: deg(60)}
    ],
    funcs: {tailAngle: pingpong(3)},
    key: 'Q'
  }
};

const Keys = {};

// Initialize the state.
copyState(initialState, state);

for (var k in Animations) {
  if (typeof Animations[k] === "object" && typeof Animations[k].key === "string")  {
    Keys[Animations[k].key] = k;
  }
}

// Keeps track of a running animation.
function AnimState(name) {
  this.name = name;
  this.animState0 = {}; // Initial state of the animation
  this.animState1 = {}; // Final state of the animation
  this.animFns = {}; // Interpolators for the animation
  this.stateId = 0;
  this.reversing = false;
}

// Advance to the next animation step.
AnimState.prototype.advance = function advance() {
  var s_start, s_end, t_length, reversing;
  var stateId = this.stateId, name = this.name;

  var states = Animations[name].states;
  if (typeof states === "undefined") {
    states = [Animations[name].start, Animations[name].end];
  }

  if (stateId + 1 >= states.length) {
    s_start = states[states.length - 1];
    s_end = states[0];
    reversing = true;
    stateId = 0;
  } else {
    if (stateId == 0) copyState(states[0], state);
    s_start = states[stateId];
    s_end = states[++stateId];
    reversing = false;
  }

  t_length = Animations[name].length || 1.0;

  this.stateId = stateId;
  this.reversing = reversing;
  this.animState0 = s_start;
  this.animState1 = s_end;
  this.animFns = Animations[name].funcs || {};
  this.time_length = t_length;
  this.time_start = clock.getElapsedTime();
  this.time_end = this.time_start + this.time_length;
}

// Update the mole's state variables by performing the interpolation.
// Returns true if and only if the whole animation sequence has finished
AnimState.prototype.update = function update() {
  var time = clock.getElapsedTime(); // t seconds passed since the clock started.
  if (jumpcut || time > this.time_end) {
    if (this.reversing) {
      this.stateId = 0;
      this.reversing = false;
      return true;
    }
    copyState(this.animState1, state);
  } else {
    var t = unlerp(time, this.time_start, this.time_end);
    evalState(this.animState0, this.animState1, this.animFns, t, state);
  }
  return false;
}

AnimState.prototype.toString = function toString() {
  return "AnimState(name=" + this.name + ", stateId="
    + this.stateId + ", reversing=" + this.reversing + ")";
}

// Start performing the next step of an animation.
function performAnimation(name) {
  var anim = Anims.find(function(a) { return a.name == name; });

  if (typeof anim === "undefined") {
    Anims.unshift(anim = new AnimState(name));
  } else {
    // move existing anim to highest priority
    Anims.splice(Anims.indexOf(anim), 1);
    Anims.unshift(anim);
  }

  anim.advance();
}

// Update the mole's body configuration
function updateBody() {
  if (Anims.length > 0) {
    for (var i = Anims.length - 1; i >= 0; i--) {
      var a = Anims[i];
      if (a.update()) Anims.splice(i, 1);
    }

    if (Anims.length == 0)
      copyState(initialState, state);

    updateMatrices();
  }
}

// LISTEN TO KEYBOARD
// Hint: Pay careful attention to how the keys already specified work!
var keyboard = new THREEx.KeyboardState();
var grid_state = false;
keyboard.domElement.addEventListener('keydown',function(event){
  if (event.repeat) return;

  if (keyboard.eventMatches(event, "Z")) {  // Z: Reveal/Hide helper grid
    grid_state = !grid_state;
    grid_state? scene.add(grid) : scene.remove(grid);
  } else if (keyboard.eventMatches(event, "0")) {    // 0: Set camera to neutral position, view reset
    camera.position.set(45,0,0);
    camera.lookAt(scene.position);
  } else if (keyboard.eventMatches(event, " ")) {
    jumpcut = !jumpcut;
  }

  for (var k in Keys) {
    if (hasKey(Keys, k)) {
      if (keyboard.eventMatches(event, k)) {
        performAnimation(Keys[k]);
        break;
      }
    }
  }
});

// SETUP UPDATE CALL-BACK
// Hint: It is useful to understand what is being updated here, the effect, and why.
function update() {
  updateBody();

  requestAnimationFrame(update);

  cubeCamera.position.set(0, 0, 0);
  cubeCamera.updateCubeMap( renderer, scene );

  renderer.render(scene,camera);
}


texture = THREE.ImageUtils.loadTexture('cubemap.jpg', THREE.UVMapping, function() {
  sphere = new THREE.Mesh( new THREE.SphereGeometry(500, 60, 40), new THREE.MeshBasicMaterial( { map: texture } ) );
  sphere.scale.x = -1;
  sphere.castShadows = sphere.receiveShadows = false;
  scene.add(sphere);

  updateMatrices();
  update();

});
